/*
* COMPONENT_NAME: ldap.clients
*
* ABSTRACT: generic program to search entries using LDAP
*
* ORIGINS: 202,27
*/ 
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <ldap.h>
#include <ldapssl.h>
#include <ldif.h>

#define DEFSEP  "="
#define DEFHOST "localhost"

#if defined( WIN32 )
#define strncasecmp strnicmp
#define prefname "ldapsearch"
#else
#define prefname "/tmp/ldapsearch"
#endif

#if defined (_HP9000) || (sunos5)
#include <unistd.h>
#endif

/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );

const char* ssl_err2string( int );

#ifndef _WIN32
#ifdef _GCC3
#include <errno.h>
#else
extern int errno;
#endif
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = (char *) "";
static char *base = NULL;
static char *ldaphost = (char *)DEFHOST;
static int ldapport = LDAP_PORT;
static char *sep = (char *) DEFSEP;
static char *mech = NULL;
static int hoplimit = 10;
static int verbose = 0;
static int allow_binary = 0;
static int ldif = 0;
static int vals2tmp = 0;
static int doit = 1;
static int ldapversion = LDAP_VERSION3;
static int DebugLevel = 0;
static int referrals = LDAP_OPT_ON;
static int attrsonly = 0;
static int scope = LDAP_SCOPE_SUBTREE;
static int deref = 0;
static FILE *fp = NULL;
static int timelimit = 0;
static int sizelimit = 0;
static int keys = 0;
static int tls = 0;
static int ssl = 0;
static int use_FIPS_mode = 0; /* 0=off 1=on */
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_label = NULL;
static char *charset = NULL;
static char *filtpattern = NULL;
static char **attrs = NULL;
static LDAP *ld = NULL;
#define FILTERSIZE 4096
char *debug_ptr = NULL;

/*Sorting*/
static int sorting = LDAP_FALSE;
static char *sortString = NULL;
static LDAPsortkey **sortKeyList = NULL;
static char sortCriticality = 'T';

/*Paging*/
static int paging = LDAP_FALSE;
static char pagingCriticality = 'T';
static int timeBetweenPages = 5;  /* default of five seconds between pages */
static struct berval *berValue = NULL;
static unsigned long totalCount = 0;
static unsigned long runningTotal = 0;
typedef struct _PageQuerySize {
   int pageSize;
   struct _PageQuerySize *next;
}
PageQuerySize;
static PageQuerySize *pageQuerySize = NULL, *curPageQuerySize = NULL, *newPageQuerySize = NULL;

/* Persistent searching */
static int pSearching = LDAP_FALSE;
static int psChangeType = 0;
static int psChangesOnly = 1;
static int psReturnEchgCtls = 1;

/*LDAPControl array for controls*/
LDAPControl **bindControls = NULL;
LDAPControl **opControls = NULL;

/* null based search requires a scope option */
static int scope_defined = LDAP_FALSE;

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends a search request to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s [-b basedn] [options] filter [attributes...]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "        basedn:     base dn for search\n");
   fprintf(stderr, "                    (optional if LDAP_BASEDN set in environment)\n");
   fprintf(stderr, "        filter:     LDAP search filter\n" );
   fprintf(stderr, "        attributes: whitespace-separated list of attributes to retrieve\n");
   fprintf(stderr, "                    (if no attribute list is specified, all are retrieved)\n");
   fprintf(stderr, "options:\n");
   fprintf(stderr, "    -a deref     how to dereference aliases (never, always, search, or find)\n");
   fprintf(stderr, "    -A           retrieve attribute names only (no values)\n");
   fprintf(stderr, "    -b base_dn   base dn for search; LDAP_BASEDN in environment is default\n");
   fprintf(stderr, "    -B           do not suppress printing of non-ASCII values\n");
   fprintf(stderr, "    -C charset   character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level     set debug level to 'level' in LDAP library\n");
   fprintf(stderr, "    -D dn        bind dn\n");
   fprintf(stderr, "    -e           display LDAP library version information and quit\n");
   fprintf(stderr, "    -f file      perform sequence of searches using filters in 'file'\n");
   fprintf(stderr, "                 \"%%s\" must be substituted for the filter\n");
   fprintf(stderr, "    -F sep       print 'sep' between attribute names and values\n");
   fprintf(stderr, "    -G realm     realm used for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -h host      LDAP server host name\n");
   fprintf(stderr, "    -i file      perform sequence of searches using filters in 'file'\n");
   fprintf(stderr, "                 \"%%s\" must be substituted for the filter\n");
   fprintf(stderr, "    -k           use server administration control on bind\n");
   fprintf(stderr, "    -K keyfile   file to use for keys\n");
   fprintf(stderr, "    -l time      time limit (in seconds) for search\n");
   fprintf(stderr, "    -L           print entries in LDIF format (-B is implied)\n");
   fprintf(stderr, "    -m mechanism perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M           manage referral objects as normal entries\n" );
   fprintf(stderr, "    -n           show what would be done but don't actually do it\n");
   fprintf(stderr, "    -N key_name  private key name to use in keyfile\n");
   fprintf(stderr, "    -o attr_type sort based on specified attribute type\n");
   fprintf(stderr, "    -O maxhops   maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port      LDAP server port number\n");
   fprintf(stderr, "    -P key_pw    keyfile password\n");
   fprintf(stderr, "    -q pagesize  query page size for paged results\n");
   fprintf(stderr, "    -R           do not automatically chase referrals\n");
   fprintf(stderr, "    -s scope     search scope (base, one, or sub)\n");
#if defined( WIN32 )
   fprintf(stderr, "    -t           write values to files\n");
#else
   fprintf(stderr, "    -t           write values to files in /tmp\n");
#endif

   fprintf(stderr, "    -T seconds   number of seconds to wait between pages for paged results\n");
   fprintf(stderr, "    -U username  user name for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -v           run in verbose mode\n");
   fprintf(stderr, "    -V version   LDAP protocol version (2 or 3; default is 3)\n");
   fprintf(stderr, "    -w password  bind password or '?' for non-echoed prompt\n");
   fprintf(stderr, "                 use backslash '\\?' to avoid matching single character filenames (UNIX only)\n" );
   fprintf(stderr, "    -x           use FIPS mode processing (SSL/TLS only)\n");
   fprintf(stderr, "    -y proxydn   set proxied id for proxied authorization operation\n");
   fprintf(stderr, "    -Y           use a secure LDAP connection (TLS)\n");
   fprintf(stderr, "    -z size      size limit (in entries) for search\n");
   fprintf(stderr, "    -Z           use a secure LDAP connection (SSL)\n");
   fprintf(stderr, "    -9 control   criticality for control option\n");
   fprintf(stderr, "                 s  sets criticality for sorting to false.\n");
   fprintf(stderr, "                 p  sets criticality for paging to false.\n");
   exit(1);
}

/*
** Print a memory error message and exit.
*/
void mem_error(const char *varName) {
   fprintf(stderr,"%s: Memory allocation error occured.\n", varName);
   exit(1);
}

int is_num(char *s) {
   int rc = 0;
   if (s && (*s == '+' || *s == '-'))
      s++;
   while (s && *s) {
      if (!isdigit(*s))
         return rc;
      s++;
   }
   return 1;
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i = 0,j = 0, port = 0, stringLen = 0;
   const char *options = "exYZnvtRMABLk?c:D:V:s:i:f:h:b:d:p:F:a:w:l:m:O:z:K:P:N:C:o:9:T:q:y:G:U:";
   LDAPVersion ver;
   extern char *optarg;
   extern int optind;
   int rc = LDAP_SUCCESS;
   int manageDsa = LDAP_FALSE;
   char *proxyDN=NULL;
   int adminControl = LDAP_FALSE;
   /* proxied authorization control request */
   int Proxy_control_requested = LDAP_FALSE;
   char *psString = NULL;
   char *tmp = NULL, *ptr = NULL;
   int psTimes = 0;
   char psChangeTypeStr[ 25 ];

   while (( i = getopt(argc, argv, options)) != EOF ) {
      switch ( i ) {
      case 'V':   /* use version 3 functions */
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Incorrect LDAP protocol version selected.\n");
            usage();
         }
         break;
      case 'c':
         if ( !(psString = strdup( optarg ) ) )
            mem_error( "psearch" );

         /* look for at minimum ps:any */
         if( strlen ( psString ) < 6 ) {
            fprintf( stderr, "Invalid argument for -C\n" );
            usage();
         } /* end if */
         pSearching = LDAP_TRUE;
         tmp = psString;
         while( ( tmp = strchr( tmp, ':' ) ) != NULL ) {
            tmp += 1;
            switch( psTimes++ ) {
            case 0:
               memset( psChangeTypeStr, '\0', 25 );
               ptr = tmp;
               while ( *ptr != ':' && *ptr != '\0' ) {
                  psChangeTypeStr[ j++ ] = *ptr;
                  ptr += 1;
               }
               if( strcmp( psChangeTypeStr, "add" ) == 0 ) {
                  psChangeType = LDAP_CHANGETYPE_ADD;
               } else if ( strcmp( psChangeTypeStr, "delete" ) == 0 ) {
                  psChangeType = LDAP_CHANGETYPE_DELETE;  
               } else if ( strcmp( psChangeTypeStr, "modify" ) == 0 ) {
                  psChangeType = LDAP_CHANGETYPE_MODIFY;
               } else if ( strcmp( psChangeTypeStr, "moddn" ) == 0 ) {
                  psChangeType = LDAP_CHANGETYPE_MODDN;
               } else if ( strcmp( psChangeTypeStr, "any" ) == 0 ) {
                  psChangeType = LDAP_CHANGETYPE_ANY;
               } else {
                  fprintf( stderr, "Invalid argument for -C\n" );
                  usage();
               }
               break;
            case 1:
               if( *tmp == '0' ) { 
                  psChangesOnly = 0;
               } else if( *tmp == '1' ) {
                  psChangesOnly = 1;
               } else {
                  fprintf( stderr, "Invalid argument for -C\n" );
                  usage();
               }
               break;               
            case 2:
               if( *tmp == '0' ) { 
                  psReturnEchgCtls = 0;
               } else if( *tmp == '1' ) {
                  psReturnEchgCtls = 1;
               } else {
                  fprintf( stderr, "Invalid argument for -C\n" );
                  usage();
               }
               break;
            default:
               fprintf( stderr, "Invalid argument for -C\n" );
               usage();               
               break;
            } /* end switch */
         } /* end while */
         break;
      case 'O':         /* set maximum referral hop count  */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         hoplimit = atoi( optarg );
         if (hoplimit < 0) {
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         break;
      case 'm':         /* use SASL bind mechanism  */
         if ( ! ( mech = strdup ( optarg )))
            mem_error ( "mech" );
         break;
      case 'n':   /* do Not do any searches */
         doit = 0;
         break;
      case 'v':   /* verbose mode */
         ++verbose;
         break;
      case 'e':
         ldap_version(&ver);
         printf ("\nSDK Version:     \t%d\n", ver.sdk_version);
         printf ("Protocol Version:\t%d\n", ver.protocol_version);
         printf ("SDK Build Level: \t%s\n", ver.sdk_build_level);
         exit(0);
      case 'd':
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect debug level.\n");
            usage();
         }
         DebugLevel = atoi( optarg );
         break;
      case 't':   /* write attribute values to /tmp files */
         ++vals2tmp;
         break;
      case 'R':   /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'M':   /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_search: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_search: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'A':   /* retrieve attribute names only -- no values */
         ++attrsonly;
         break;
      case 'L':   /* print entries in LDIF format */
         ++ldif;
         break;
      case 'B':   /* allow binary values to be printed */
         ++allow_binary;
         break;
      case 's':   /* search scope */
         if ((strlen( optarg ) == 4) && ( strncasecmp( optarg, "base", 4 )) == 0 ) {
            scope = LDAP_SCOPE_BASE;
         } else if ((strlen( optarg ) == 3) &&( strncasecmp( optarg, "one", 3 )) == 0 ) {
            scope = LDAP_SCOPE_ONELEVEL;
         } else if ((strlen( optarg ) == 3) &&( strncasecmp( optarg, "sub", 3 )) == 0 ) {
            scope = LDAP_SCOPE_SUBTREE;
         } else {
            fprintf( stderr, "scope should be base, one, or sub\n" );
            usage();
         }
         scope_defined = LDAP_TRUE ;
         break;
      case 'a':   /* set alias deref option */
         if ( strncasecmp( optarg, "never", 5 ) == 0 ) {
            deref = LDAP_DEREF_NEVER;
         } else if ( strncasecmp( optarg, "search", 5 ) == 0 ) {
            deref = LDAP_DEREF_SEARCHING;
         } else if ( strncasecmp( optarg, "find", 4 ) == 0 ) {
            deref = LDAP_DEREF_FINDING;
         } else if ( strncasecmp( optarg, "always", 6 ) == 0 ) {
            deref = LDAP_DEREF_ALWAYS;
         } else {
            fprintf( stderr, "alias deref should be never, search, find, or always\n" );
            usage();
         }
         break;
      case 'F':   /* field separator */
         if ( ! ( sep = strdup ( optarg )))
            mem_error( "sep" );
         break;
      case 'i':
      case 'f':   /* input file */
         if (optarg[0] == '-' && optarg[1] == '\0' )
            fp = stdin;
         else if (( fp = fopen( optarg, "r" )) == NULL ) {
            perror( optarg );
            exit( 1 );
         }
         break;
      case 'h':   /* ldap host */
         if ( ! ( ldaphost = strdup ( optarg )))
            mem_error ( "ldaphost" );
         break;
      case 'b':   /* searchbase */
         if ( ! ( base = strdup ( optarg )))
            mem_error ( "base" );
         break;
      case 'D':   /* bind DN */
         if ( ! ( binddn = strdup ( optarg )))
            mem_error ( "binddn" );
         break;
      case 'p':   /* ldap port */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         ldapport = atoi( optarg );
         if (ldapport <= 0) {
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         port = 1;
         break;
      case 'w':   /* bind password */
         if (optarg && optarg[0] == '?') {
            passwd = getPassword();

         } else
            if ( ! ( passwd = strdup ( optarg )))
               mem_error ( "passwd" );
         break;
      case 'l':   /* time limit */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect time limit.\n");
            usage();
         }
         timelimit = atoi( optarg );
         if (timelimit < 0) {
            fprintf(stderr, "Incorrect time limit.\n");
            usage();
         }
         break;
      case 'z':   /* size limit */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect size limit.\n");
            usage();
         }
         sizelimit = atoi( optarg );
         if (sizelimit < 0 ) {
            fprintf(stderr, "Incorrect size limit.\n");
            usage();
         }
         break;
      case 'K':
         if ( ! ( keyfile = strdup ( optarg )))
            mem_error ( "keyfile" );
         keys = 1 ;
         break;
      case 'P':
         if ( ! ( keyfile_pw = strdup ( optarg )))
            mem_error ( "keyfile_pw" );
         break;
      case 'Y':
         if (ssl) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         tls = 1;
         break;
      case 'x':
         use_FIPS_mode = 1;
         break;
      case 'Z':
         if (tls) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         ssl = 1;
         break;
      case 'N':
         if ( ! ( keyfile_label = strdup ( optarg )))
            mem_error ( "keyfile_label" );
         break;
      case 'C':
         if (!(charset = strdup(optarg)))
            mem_error("charset");
         break;
      case 'o':
         sorting = LDAP_TRUE;
         /* copy key to sort string, prefix space as string delimeter */
         if (sortString == NULL) {
            stringLen = strlen(optarg) + 1; /* 1 for beginning blank */
            sortString = (char *) malloc(stringLen + 1); /* 1 for end of string */
            if (sortString == NULL)
               mem_error("sortString");
            memset(sortString, ' ', 1);
            memcpy(sortString + 1, optarg, strlen(optarg));
            memset(sortString + 1 + strlen(optarg), '\0', 1);
         } else {
            char *newSortString = (char *) realloc(sortString, (stringLen + strlen(optarg) + 1 + 1));
            sortString = newSortString;
            memset(sortString + stringLen, ' ', 1);
            memcpy(sortString + stringLen + 1, optarg, strlen(optarg));
            stringLen += strlen(optarg) + 1;
            memset(sortString + stringLen, '\0', 1);
         }
         break;
      case 'T':   /* time, in seconds, between pages of search results */
         paging = LDAP_TRUE;
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect time interval.\n");
            usage();
         }
         timeBetweenPages = atoi( optarg );
         if (timeBetweenPages < 0) {
            fprintf(stderr, "Incorrect time interval.\n");
            usage();
         }
         break;
      case 'q':
         paging = LDAP_TRUE;
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect page size.\n");
            usage();
         }
         newPageQuerySize = (PageQuerySize *) malloc(sizeof(PageQuerySize));
         if (newPageQuerySize == NULL) {
            mem_error("pageQuerySize");
            break;
         }
         newPageQuerySize->pageSize = atoi(optarg);
         if (newPageQuerySize->pageSize < 0) {
            fprintf(stderr, "Incorrect page size.\n");
            usage();
         }
         newPageQuerySize->next = NULL;
         if (curPageQuerySize == NULL) {
            pageQuerySize = curPageQuerySize = newPageQuerySize;
         } else {
            curPageQuerySize->next = newPageQuerySize;
            curPageQuerySize = newPageQuerySize;
         }
         break;
      case '9':
         if ((strcmp(optarg, "S") == 0) || (strcmp(optarg, "s") == 0))
            sortCriticality = 'F';
         else if ((strcmp(optarg, "P") == 0) || (strcmp(optarg, "p") == 0))
            pagingCriticality = 'F';
         else {
            fprintf(stderr, "Incorrect control criticality.\n");
            usage();
         }
         break;
      case 'k':  /* admin control */
         adminControl = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'y':   /* Proxied authorization control */
            rc=ldap_proxy_dn_prefix(&proxyDN,optarg);
            if (rc == LDAP_SUCCESS) {
                stringLen=strlen(proxyDN);
                rc = ldap_add_control (LDAP_PROXY_CONTROL_OID,stringLen,
                      proxyDN,1,&opControls);
               free(proxyDN);
              }
          if (rc != LDAP_SUCCESS) {
              fprintf(stderr, "Incorrect Proxied DN specified. \n");
              usage();
               }
           else  Proxy_control_requested=LDAP_TRUE;
          break;
      case 'U':  /* DIGEST-MD5 Username */
         if ( !(digest_username = strdup( optarg )) ) {
            mem_error( "Username" );
         }
         break;
      case 'G':  /* DIGEST-MD5 Realm */
         if ( !(digest_realm = strdup( optarg )) ) {
            mem_error( "realm" );
         }
         break;
      case '?':
      default:
         usage();
      }
   }

   if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
      ssl = 1;                  /* Use SSL by default */

   if (( manageDsa | sorting | paging | Proxy_control_requested | adminControl) && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M, -o, -q, -k and -T options require at least LDAP version 3.\n");
      if (manageDsa)
         fprintf(stderr, "-M ignored.\n");
      if (sorting)
         fprintf(stderr, "-o ignored.\n");
      if (paging)
         fprintf(stderr, "-T and -q ignored.\n");
      if (adminControl)
         fprintf(stderr, "-k ignored.\n");
      if (Proxy_control_requested)
          fprintf(stderr, "-y ignored.\n");
      fprintf(stderr, "\n\n");
      manageDsa = LDAP_FALSE;
      sorting = LDAP_FALSE;
      paging = LDAP_FALSE;
   }

   if (argc - optind < 1)
      usage();

   if (!(filtpattern = strdup(argv[optind])))
      mem_error ("filtpattern");

   if (argv[optind + 1] == NULL)
      attrs = NULL;
   else
      attrs = &argv[ optind + 1 ];

   if (base == NULL) {
      base = getenv( "LDAP_BASEDN" );
      if (base != NULL)
         base = strdup(base);
   }

   /* To prevent user from inadvertently doing a
      null based subtree search and consuming a
      lot of server's resources, the scope option
      is required for a null based search request. */
   if ( (base == NULL || *base == '\0') && !scope_defined ) {
      fprintf(stderr, "scope is required for a null based search.\n");
      usage();
   }

   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT; /* Use default SSL port if none specified */

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );

   /*Set up the controls*/
   /*Add the password policy controls*/
   if (ldapversion == LDAP_VERSION3) {
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &bindControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &opControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
   }
}


/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   fputs( "Enter password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}


/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
         *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/

void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **bind_client_controls = NULL;

   /* Set FIPS mode processing if requested for TLS and SSL. */
   if (( tls == 1 || ssl == 1 ) && use_FIPS_mode == 1 ) {
      if (verbose)
	 printf( "ldap_ssl_set_fips_mode_np(1)\n");
      
      rc = ldap_ssl_set_fips_mode_np( 1 ); 
      if (rc != LDAP_SUCCESS) {
	 fprintf( stderr,
		  "ldap_ssl_set_fips_mode_np failed\n" );
	 exit( rc ) ;
      }
   }

   /* call ldap_ssl_client_init if V3 and SSL */
   if ((tls || ssl) && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         fprintf( stderr, "%s\n", ssl_err2string( failureReasonCode ));
         exit( 1 ) ;
      }
   }

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport,
                   ((keyfile_label) ? keyfile_label : "NULL"));
         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_label ) ;
         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed with invalid parm\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport);
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
            fprintf( stderr, "ldap_init failed with invalid parm\n" ) ;
            perror("ldap_init");
            exit(1);
         }
      }
   } else {
      if ( verbose )
         printf( "ldap_open( %s, %d )\n",
                 ((ldaphost) ? ldaphost : "NULL"), ldapport );
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *)&timelimit);
      ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *)&sizelimit);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )(long)DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )(long)deref);
      ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *)(long)timelimit);
      ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *)(long)sizelimit);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)(long)hoplimit);
   }
   /* don't chase referrals during the bind */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_OFF);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }

   /* call ldap_ssl_start if V2 and SSL */
   if ( (tls || ssl) && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_label) ? keyfile_label : "NULL"));

      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_label);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr, "ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   if (tls) {                   /* Start the TLS session */
      if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
         fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_start_tls_s_np: %s\n", errormsg);
         exit(rc);
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {
      if ( ! mech ) /* Use simple bind */
      {
	 ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, bindControls, NULL, &msgidp);
         rc2 = rc; /* Store this result code for later user. */
         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, NULL,
                                    NULL, &returnedControls, 1 );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               ldap_get_option( ld, LDAP_OPT_EXT_GSS_ERR, &failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Major Status: %i\n", failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Minor Status: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else /* other SASL mechanisms */ {
            if (strcmp(mech,LDAP_MECHANISM_DIGEST_MD5) == 0) {
               if (digest_username) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) , digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
               if (digest_realm) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) , digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
            }

            ber.bv_len = strlen ( passwd );
            ber.bv_val = passwd;
            rc = ldap_sasl_bind_s (ld, binddn, mech, &ber,
                                   bindControls, bind_client_controls,
                                   &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               exit ( rc );
            }
         }

         /* Get any controls sent by the server during the sasl_bind_s */
         rc2 = rc;

         if ( rc == LDAP_SUCCESS ) {
            rc = ldap_get_bind_controls( ld, &returnedControls );
         }
      }

      /*Free the client controls*/
      ldap_controls_free(bind_client_controls);

      if (rc == LDAP_SUCCESS) {
         if (returnedControls != NULL) {
            ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
         }
      }
      if (rc2 != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc2);
      } else if (rc != LDAP_SUCCESS) {
         errormsg = ldap_err2string(error);
      } else if (errcode != LDAP_SUCCESS) {
         errormsg = ldap_err2string(errcode);
      }
      if (controlerr) {
         errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
      }
      if (controlwarn) {
         warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
      }
      if ((errormsgctrl != NULL)  || (warnmsgctrl != NULL)) {
         fprintf(stderr,"ldap_simple_bind:  ");
         if (errormsg != NULL) {
            fprintf(stderr,errormsg);
            fprintf(stderr," --- ");
         }
         if (errormsgctrl != NULL) {
            fprintf(stderr,errormsgctrl);
            fprintf(stderr,"\n");
         }
         if (warnmsgctrl != NULL) {
            fprintf(stderr,warnmsgctrl, controlres);
            fprintf(stderr,"\n");
         }
      } else if (errormsg != NULL) {
         fprintf(stderr,"ldap_simple_bind: %s \n", errormsg);
      }
      if (rc2 != LDAP_SUCCESS) {
         /*
         ** ldap_get_exterror only valuable for ssl communication.
         ** In this example, for LDAP v3, the bind is the first
         ** instance in which communication actually flows to the
         ** server.  So, if there is an ssl configuration error or
         ** other ssl problem, this will be the first instance where
         ** it will be detected.
         */
         if (ssl) {
            fprintf( stderr, "Attempted communication over SSL.\n");
            fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
         }
         exit( rc2 );
      } else if (errcode != LDAP_SUCCESS) {
         exit( errcode );
      }

   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }

   /* during the operation, use user setting to chase referrals */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )(long)referrals);
}

/* ssl_err2string returns a string describing the error associated with
** the given error code from an SSL related call.
*/
const char * ssl_err2string( int ssl_error ) {
   switch ( ssl_error ) {
   case LDAP_SSL_EXTENDED_OK:
      return LDAP_SSL_EXTENDED_OK_STR;
   case LDAP_SSL_EXTENDED_INVALID_HANDLE:
      return LDAP_SSL_EXTENDED_INVALID_HANDLE_STR;
   case LDAP_SSL_EXTENDED_API_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_API_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_INTERNAL_ERROR:
      return LDAP_SSL_EXTENDED_INTERNAL_ERROR_STR;
   case LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE:
      return LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE_STR;
   case LDAP_SSL_EXTENDED_INVALID_STATE:
      return LDAP_SSL_EXTENDED_INVALID_STATE_STR;
   case LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND:
      return LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND_STR;
   case LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION:
      return LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_CRYPTO:
      return LDAP_SSL_EXTENDED_ERROR_CRYPTO_STR;
   case LDAP_SSL_EXTENDED_ERROR_ASN:
      return LDAP_SSL_EXTENDED_ERROR_ASN_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR:
      return LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR_STR;
   case LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR:
      return LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR:
      return LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT:
      return LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL_STR;
   case LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD:
      return LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED:
      return LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB:
      return LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB_STR;
   case LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR:
      return LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR_STR;
   case LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR:
      return LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR_STR;
   case LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR:
      return LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_INDEX:
      return LDAP_SSL_EXTENDED_ERROR_BAD_INDEX_STR;
   case LDAP_SSL_EXTENDED_CLOSE_FAILED:
      return LDAP_SSL_EXTENDED_CLOSE_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_DATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_DATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS:
      return LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_IO:
      return LDAP_SSL_EXTENDED_ERROR_IO_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MAC:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MAC_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_PEER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_PEER_STR;
   case LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED:
      return LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED_STR;
   case LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED:
      return LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED:
      return LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY:
      return LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER:
      return LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER_STR;
   case LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE:
      return LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE_STR;
   case LDAP_SSL_EXTENDED_WOULD_BLOCK:
      return LDAP_SSL_EXTENDED_WOULD_BLOCK_STR;
   case LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3:
      return LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3_STR;
   case LDAP_SSL_EXTENDED_MISC_INVALID_ID:
      return LDAP_SSL_EXTENDED_MISC_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE_STR;
   case LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING:
      return LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING_STR;
   default:
      return "Unknown SSL error";
   }
}


/*
** Write an ldif style value to standard output.
*/
int write_ldif_value(const char *type, const char *value, unsigned long vallen) {
   char *ldif;

   if ((ldif = ldif_type_and_value(type, (char*) value, (int)vallen)) == NULL)
      return -1;
   fputs(ldif, stdout);
   free(ldif);
   return 0;
}

/*
** Write an ldif value or binary value to standard output.
*/
int write_ldif_value_or_bvalue(char *type, char *value, unsigned long vallen,
                               char *bvalue, unsigned long bvallen) {
   char *ldif;

   ldif = ldif_type_and_value_or_bvalue(type, value, (int)vallen,
                                        bvalue, (int)bvallen);
   if (ldif == NULL)
      return -1;
   fputs(ldif, stdout);
   free(ldif);
   return 0;
}

/*
** Write entry to temp file.
*/
void write_tmp_file(char *attr, struct berval *bval) {
   char postfname[128], tmpfname[128];
   FILE *tmpfp = NULL;

   sprintf(postfname, "-%s-XXXXXX", attr);
   strcpy(tmpfname, prefname);
   strcat(tmpfname, postfname);

   /* CMVC d91505 */    
   #if defined LINUX
   mkstemp(tmpfname); /* mktemp was flagged as dangerous for Linux and the use of mkstemp was recommended by gcc. Hence the change*/
   #else
   mktemp(tmpfname); 
   #endif

   if ((tmpfp = fopen( tmpfname, "wb")) == NULL) {
      perror(tmpfname);
      return ;
   }
   if (fwrite(bval->bv_val, bval->bv_len, 1, tmpfp) == 0)
      perror(tmpfname);
   else if (ldif)
      write_ldif_value(attr, tmpfname, strlen(tmpfname));
   else
      printf("%s%s%s\n", attr, sep, tmpfname);
   fclose(tmpfp);
}

/*
** Output attribute information.
*/
void print_value(char *attr, struct berval *bval, char *val) {
   int notascii = 0, i;

   if (ldif)
      write_ldif_value_or_bvalue(attr, val, bval->bv_len,
                                 bval->bv_val, bval->bv_len);
   else {
      /* if charset is set then check for ascii is invalid. */
      if (val == NULL)
         notascii = 1;
      else if (charset == NULL) {
         if (strlen(val) == bval->bv_len) {
            for (i = 0; i < bval->bv_len; i++) {
               if (!isascii(val[i])) {
                  notascii = 1;
                  break;
               }
            }
         } else
            notascii = 1;
      }

      printf("%s%s", attr, sep);
      if (notascii)
         printf("%s\n", (allow_binary) ? bval->bv_val : "NOT ASCII");
      else
         printf("%s\n", val);
   }
}

/*
** Output a search result.
*/
void print_entry(LDAPMessage *entry) {
   char *attr, *dn;
   int i;
   BerElement *ber;
   struct berval **bvals;
   char **vals;

   /* write dn */
   dn = ldap_get_dn( ld, entry );
   if (ldif)
      write_ldif_value( "dn", dn, strlen( dn ));
   else
      printf( "%s\n", dn );

   ldap_memfree( dn );

   /* write attributes */
   for (attr = ldap_first_attribute( ld, entry, &ber ); attr != NULL;
        attr = ldap_next_attribute( ld, entry, ber )) {
      /* attribute name only */
      if (attrsonly) {
         if (ldif)
            write_ldif_value(attr, "", 0);
         else
            printf("%s\n", attr);

         /* attribute name and value */
      } else if ((bvals = ldap_get_values_len(ld, entry, attr)) != NULL) {
         vals = ldap_get_values( ld, entry, attr);
         for ( i = 0; bvals[i] != NULL; i++) {
            if (vals2tmp)
               write_tmp_file(attr, bvals[i]);
            else
               print_value(attr, bvals[i],
                           (vals == NULL) ? NULL : vals[i]);
         }
         ldap_value_free_len(bvals);
         ldap_value_free(vals);
      }
      ldap_memfree(attr);
   }

   if ( ber != NULL )
      ldap_ber_free ( ber );

}

static void printPsearchChangeInfo( int change ) {
   switch ( change ) {
   case LDAP_CHANGETYPE_ADD:
      fprintf( stderr, "persistentSearch-changeType: %s\n", "add");
      break;
   case LDAP_CHANGETYPE_DELETE:
      fprintf( stderr, "persistentSearch-changeType: %s\n", "delete" );
      break;
   case LDAP_CHANGETYPE_MODIFY:
      fprintf( stderr, "persistentSearch-changeType: %s\n", "modify" );
      break;
   case LDAP_CHANGETYPE_MODDN:
      fprintf( stderr, "persistentSearch-changeType: %s\n", "moddn" );
      break;
   default:
      break;
   }
}


/*
** Run the search.
*/
int dosearch(char *value) {
   char filter[ FILTERSIZE ], **val;
   char **referrals = NULL;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *matched, *errmsg;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   int msgid, rc, rc2, first, matches, references, errcode, i;
   LDAPMessage *res, *e;
   LDAPControl **returnedControls = NULL;
   char *keyInError = NULL;
   unsigned long sortRC = LDAP_SUCCESS;
   int changeType = 0;
   char *prevDN = NULL;
   long changeNum = 0;
   int changeNumPresent = 0;
   
   /* initialize filter */
   if ( value )
      sprintf(filter, filtpattern, value);
   else
      strncpy(filter, filtpattern, FILTERSIZE - 1);

   /* verbose */
   if (verbose)
      printf( "filter is: (%s)\n", filter );
   if (!doit)
      return LDAP_SUCCESS;

   /* send search request */
   if (ldapversion == LDAP_VERSION3) {
      rc = ldap_search_ext( ld, base, scope, filter, attrs, attrsonly,
                            opControls,
                            NULL, NULL, sizelimit, &msgid);
      if ( rc != LDAP_SUCCESS ) {
         fprintf ( stderr, "ldap_search_ext: %s\n", ldap_err2string (rc));
         return ldap_get_errno(ld);
      }

   } else {
      msgid = ldap_search( ld, base, scope, filter, attrs, attrsonly );
      if ( msgid == -1) {
         rc = ldap_get_errno(ld);
         fprintf ( stderr, "ldap_search: %s\n", ldap_err2string (rc));
         return rc;
      }
   }

   /* disable translation if we're outputing ldif format */
   if ((ldif) && (charset != NULL))
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_OFF);

   /* get and print search results */
   matches = 0;
   references = 0;
   first = 1;
   while (1) {
      rc = ldap_result( ld, msgid, LDAP_MSG_ONE, NULL, &res );

      /* search result */
      if ( rc == LDAP_RES_SEARCH_ENTRY ) {

         matches++;
         e = ldap_first_entry( ld, res );
         if ( !first )
            putchar( '\n' );
         else
            first = 0;
         print_entry(e);
         
         /* psearch - parse the control for each entry returned
            and print the appropriate information */
         if( pSearching ) {
            rc = ldap_get_entry_controls( ld, e, &returnedControls );
            
            if( rc == LDAP_SUCCESS ) {
               if ( ( returnedControls != NULL ) && ( ldapversion == LDAP_VERSION3 ) ) {
                  rc = ldap_parse_entrychange_control( ld, returnedControls,
                                                       &changeType, &prevDN,
                                                       &changeNumPresent, 
                                                       &changeNum );
                  if ( rc == LDAP_SUCCESS ) {
                     /* print the change type */
                     printPsearchChangeInfo( changeType );
                     /* If there was a previous dn (modrdn), print it */
                     if( prevDN )
                        fprintf( stderr, "persistentSearch-previousDN: %s\n",
                                 prevDN );
                     /* if there was a changenum (changelog), print it */
                     if( changeNumPresent > 0 )
                        fprintf ( stderr,"persistentSearch-changeNumber: %d\n",
                                  changeNum );
                  }
               }
            }
         } else {
            ldap_msgfree( res );
         }
         /* search reference */
      } else if ( rc == LDAP_RES_SEARCH_REFERENCE ) {
         references++;
         ldap_parse_reference( ld, res, &referrals, NULL, 1 );
         if (referrals != NULL) {
            for (i = 0; referrals[i] != NULL; i++)
               fprintf(stderr, "Unfollowed search reference: %s\n",
                       referrals[i]);
            fflush(stderr);
            ldap_value_free(referrals);
            referrals = NULL;
         }

         /* done */
      } else
         break;
   }

   /* check result code */
   if ( rc == -1 ) {
      errormsg = ldap_err2string(ldap_get_errno(ld));
      fprintf(stderr,"ldap_result: %s \n", errormsg);
      return rc;
   }

   rc = ldap_parse_result( ld, res, &errcode, &matched, &errmsg,
                           &referrals, &returnedControls, 1 );

   if (rc == LDAP_SUCCESS) {
      if ((returnedControls != NULL) && ( ldapversion == LDAP_VERSION3)){
         ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
      }
      if ( errcode != LDAP_SUCCESS ) {
         int printed;
         printed = 0;
         rc = errcode;  /* capture return code */

         errormsg = ldap_err2string(errcode);

         if (controlerr) {
            errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
         }
         if (controlwarn) {
            warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
         }
         if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
            printed = 1;
            fprintf( stderr, "ldap_search:  ");
            if (errormsg != NULL) {
               fprintf(stderr, errormsg);
               fprintf(stderr, " --- \n");
            }
            if (errormsgctrl != NULL) {
               fprintf(stderr, errormsgctrl);
               fprintf(stderr, "\n");
            }
            if (warnmsgctrl != NULL) {
               fprintf(stderr, warnmsgctrl, controlres);
               fprintf(stderr, "\n");
            }
         }

         if ( matched != NULL )
            if ( *matched != '\0' ) {
               printed = 1;
               fprintf( stderr, "ldap_search: %s\n", errormsg );
               fprintf( stderr, "ldap_search: matched: %s\n", matched );
            }
         if ( errmsg != NULL )
            if ( *errmsg != '\0' ) {
               printed = 1;
               fprintf( stderr, "ldap_search: %s\n", errormsg );
               fprintf( stderr, "ldap_search: additional info: %s\n", errmsg );
            }
         if (printed == 0 ) /* we didnt visit any routines that printed so this should now print*/
            fprintf( stderr, "ldap_search: %s\n", errormsg );

      }
   } else
      fprintf( stderr, "ldap_search: error parsing result: %d, %s\n",
               rc, ldap_err2string( rc ) );

   if ( tls )
      ldap_stop_tls_s_np ( ld );

   if ( matched )
      ldap_memfree ( matched );

   if ( errmsg )
      ldap_memfree ( errmsg );

   if ( referrals != NULL ) {
      int i;
      for ( i = 0; referrals[i] != NULL; i++) {
         fprintf( stderr, "%s %s\n",
                  (i == 0) ? "Unfollowed referral:" :
                  "                    ",
                  referrals[i]);
      }
      ldap_value_free( referrals );
      referrals = NULL;
   }

   /* decode page control */
   if (berValue) {
      ber_bvfree(berValue);
      berValue = NULL;
   }
   if ((returnedControls != NULL) && (pageQuerySize != NULL) && (ldapversion == LDAP_VERSION3)) {
      rc = ldap_parse_page_control(ld, returnedControls, &totalCount, &berValue);

      if (rc != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_search: error parsing page result: %d, %s\n",
                  rc, ldap_err2string( rc ) );
      }
   }

   if ((pageQuerySize != NULL) &&
       (ldapversion == LDAP_VERSION3) &&
       (rc == LDAP_SUCCESS)) {
     runningTotal += matches;
     if (verbose) {
       printf("%d total entries have been returned\n", runningTotal);
     }
   }

   /* process result controls  */
   if ((returnedControls != NULL) && (sorting) && (ldapversion == LDAP_VERSION3)) {
      /* decode sort control */
      rc = ldap_parse_sort_control(ld, returnedControls, &sortRC, &keyInError);

      if (rc != LDAP_SUCCESS) {
         /* don't print error message if this is the last page of paged results */
         /* request, we don't expect any sort control in this case    */
         fprintf( stderr, "ldap_search: error parsing sort result: %d, %s\n",
                  rc, ldap_err2string( rc ) );
      } else {
         /* display results to user */
         switch (sortRC) {
         case LDAP_SUCCESS:
            if (verbose)
               printf("Sorted search control successful\n");
            break;
         case LDAP_UNWILLING_TO_PERFORM:
         case LDAP_NO_SUCH_ATTRIBUTE:
         case LDAP_INAPPROPRIATE_MATCHING:
         case LDAP_ADMIN_LIMIT_EXCEEDED:
            if (keyInError) {
               fprintf( stderr, "Sorted search control error:  %s, %s\n",
                        ldap_err2string(sortRC), keyInError);
            } else {
               fprintf( stderr, "Sorted search control error:  %s\n",
                        ldap_err2string(sortRC));
            }
            break;
         case LDAP_SIZELIMIT_EXCEEDED:
            fprintf( stderr, "Sorted search control error:  %s\n",
                     ldap_err2string(LDAP_ADMIN_LIMIT_EXCEEDED));
            break;
         case LDAP_OPERATIONS_ERROR:
         case LDAP_BUSY:
         case LDAP_INSUFFICIENT_ACCESS:
         case LDAP_UNAVAILABLE_CRITICAL_EXTENSION:
         default:
            fprintf( stderr, "Sorted search control error:  %s\n", ldap_err2string(sortRC));
            break;
         }
      }
   }
   fflush( stderr );
   if ( verbose ) {
      printf( "%d matches\n", matches );
      if (references > 0)
         printf( "%d unfollowed references\n", references );
   }

   /* re-enable translation if we previously turned it off */
   if ((ldif) && (charset != NULL))
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);

   /* clean up  */
   ldap_controls_free(returnedControls);
   returnedControls = NULL;
   if (keyInError)
      free(keyInError);
   keyInError = NULL;

   return rc ;
}


int main( int argc, char **argv ) {
   char line[ FILTERSIZE ];
   int rc = LDAP_SUCCESS, rc2= LDAP_SUCCESS, i, first, ii, firstTime = 1;
   FILE * cf_fd;
   LDAPControl *sortControl = NULL, *pageControl = NULL, *psControl = NULL;

   /* Strip off any path info on program name */
#if defined( WIN32 ) || defined( _WIN32 )

   if ((prog = strrchr(argv[0], '\\')) != NULL)
      ++prog;
   else
      prog = argv[0];
#else

   if (prog = strrchr(argv[0], '/'))
      ++prog;
   else
      prog = argv[0];
#endif

   /* parse command line arguments */
   parse_arguments(argc, argv);

   /* connect to server */
   if (doit)
      connect_to_server();

   /* verbose */
   if ( verbose ) {
      printf( "filter pattern: %s\nreturning: ", filtpattern );
      if ( attrs == NULL )
         printf( "ALL" );
      else {
         for ( i = 0; attrs[ i ] != NULL; ++i )
            printf( "%s ", attrs[ i ] );
      }
      putchar( '\n' );
   }

   /* sorted search */
   if (sortString && sorting ){
      if ((rc = ldap_create_sort_keylist(&sortKeyList, sortString)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_search: ldap_create_sort_keylist: %d, %s\n",
                  rc, ldap_err2string( rc ) );
      } else {
         if ((rc = ldap_create_sort_control(ld, sortKeyList, sortCriticality, &sortControl)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_search: ldap_create_sort_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
         } else {
            if ((rc = ldap_insert_control( sortControl, &opControls)) != LDAP_SUCCESS)
               fprintf( stderr, "ldap_search: ldap_insert_control: %d, %s\n",
                        rc, ldap_err2string( rc ) );
         }
         ldap_free_sort_keylist(sortKeyList);
         if (sortString)
            free (sortString);
         sortString = NULL;
      }
   }

   /* psearch */
   if( pSearching ) {
      rc = ldap_create_persistentsearch_control( ld, 
                                                 psChangeType,
                                                 psChangesOnly, 
                                                 psReturnEchgCtls, 
                                                 0, 
                                                 &psControl );
      if( rc != LDAP_SUCCESS )
         fprintf( stderr, "ldap_search: ldap_create_persistentsearch_control: %s\n",
                  ldap_err2string( rc ) );
      
      if ((rc = ldap_insert_control( psControl, &opControls)) != LDAP_SUCCESS)
         fprintf( stderr, "ldap_search: ldap_insert_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );

   }
   
   /* do the search */
   if (rc == LDAP_SUCCESS) {
      if (fp == NULL) {
         if (pageQuerySize && paging) {
            curPageQuerySize = pageQuerySize;
            while (curPageQuerySize && (firstTime || (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0)))) {
               if ((rc = ldap_create_page_control(ld, curPageQuerySize->pageSize, berValue, pagingCriticality, &pageControl)) != LDAP_SUCCESS) {
                  fprintf( stderr, "ldap_search: ldap_create_page_control: %d, %s\n",
                           rc, ldap_err2string( rc ) );
               } else {
                  if ((rc = ldap_insert_control( pageControl, &opControls)) != LDAP_SUCCESS)
                     fprintf( stderr, "ldap_search: ldap_insert_control: %d, %s\n",
                              rc, ldap_err2string( rc ) );
               }
               rc = dosearch(NULL);
               rc2 = ldap_remove_control(pageControl, &opControls, LDAP_TRUE);
               if (rc)
                  break;
               if (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0))
#ifdef _WIN32

                  Sleep(timeBetweenPages*1000);
#else

               sleep(timeBetweenPages);
#endif

               printf( "\n" );
               if (firstTime == 1)
                  firstTime = 0;
               if (curPageQuerySize->pageSize == 0)
                  break;
               if (curPageQuerySize->next != NULL)
                  curPageQuerySize = curPageQuerySize->next;
            }
         } else {
            rc = dosearch(NULL);
         }
      } else {
         rc = 0;
         first = 1;
         while ( rc == 0 && fgets( line, sizeof( line ), fp ) != NULL ) {
            if (!first)
               putchar( '\n' );
            else
               first = 0;

            /* strip newlines and carriage-returns if exists */
            i = strlen(line);
            if (i > 0 && (line[i - 1] == '\r' || line[i - 1] == '\n'))
               line[--i] = '\0';

            /* skip blank lines and comment lines */
            if (*line == '\0' ||
                *line == '\b' ||
                *line == '\n' ||
                *line == '\t' ||
                *line == ' ' ||
                strstr(line, "/*") ||
                strstr(line, "//"))
               continue;

            /* loop through the pages for this search filter */
            if (pageQuerySize && paging) {
               curPageQuerySize = pageQuerySize;
               while (curPageQuerySize && (firstTime || (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0)))) {
                  if ((rc = ldap_create_page_control(ld, curPageQuerySize->pageSize, berValue, pagingCriticality, &pageControl)) != LDAP_SUCCESS) {
                     fprintf( stderr, "ldap_search: ldap_create_page_control: %d, %s\n",
                              rc, ldap_err2string( rc ) );
                  } else {
                     if ((rc = ldap_insert_control( pageControl, &opControls)) != LDAP_SUCCESS)
                        fprintf( stderr, "ldap_search: ldap_insert_control: %d, %s\n",
                                 rc, ldap_err2string( rc ) );
                  }

                  rc = dosearch(line);
                  rc2 = ldap_remove_control(pageControl, &opControls, LDAP_TRUE);
                  if (rc)
                     break;
                  if (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0))
#ifdef _WIN32

                     Sleep(timeBetweenPages*1000);
#else

                  sleep(timeBetweenPages);
#endif

                  printf( "\n" );
                  if (firstTime == 1)
                     firstTime = 0;
                  if (curPageQuerySize->pageSize == 0)
                     break;
                  if (curPageQuerySize->next != NULL)
                     curPageQuerySize = curPageQuerySize->next;
               }
            } else {
               rc = dosearch(line);
            }
         }
         if ( fp != stdin )
            fclose( fp );
      }
   }

   /* unbind & exit */
   if (doit && !tls)
      ldap_unbind( ld );

   /* clean up */
   if (opControls)
      ldap_controls_free(opControls);
   if (bindControls)
      ldap_controls_free(bindControls);
   while (pageQuerySize) {
      curPageQuerySize = pageQuerySize->next;
      free(pageQuerySize);
      pageQuerySize = curPageQuerySize;
   }
   pageQuerySize = NULL;
   if (strcmp(passwd,""))
      free(passwd);
   if (base)
      free(base);
   if (filtpattern)
      free(filtpattern);
   if (binddn)
      free(binddn);
   if (berValue)
      ber_bvfree(berValue);

   exit(rc);
}

/* end of ldapsearch.c */
